unit Render;

interface

uses
  ControlPoint, ControlPoints, Windows, Graphics, dialogs;

type
  TOnProgress = procedure(prog: double) of object;

type
  TRenderer= class
  private
    Fcp: TControlPoint;
    Fcps: TcontrolPoints;
    bm: TBitmap;

    oversample: integer;
    filter_width: integer;
    filter: array of array of double;

    image_Width: integer;
    image_Height: integer;
    BucketWidth: integer;
    BucketHeight: integer;
    BucketSize: integer;
    gutter_width : integer;

    sample_density: double;

    buckets: array of array of integer;
    accumulate: array of array of integer;

    cmap: array[0..255,0..3] of integer;

    bg: array[0..2] of double;
    vib_gam_n: integer;
    vibrancy: double;
    gamma: double;

    bounds: array[0..3] of double;
    size: array[0..1] of double;
    ppux, ppuy: double;
    bStop: boolean;

    temporal_filter: array of double;
    temporal_deltas: array of double;

    bUseAcculationBuffer: boolean;

    FOnProgress: TOnProgress;

    procedure SetOnProgress(const Value: TOnProgress);

    procedure CreateFilter;
    procedure NormalizeFilter;

    procedure InitValues;
    procedure InitBuffers;
    procedure InitBitmap(w: integer = 0; h: integer = 0);
    procedure InitTemporalData;
    procedure ClearBuffers;
    procedure ClearBuckets;
    procedure CreateCMap;
    procedure CreateCamera;

    procedure FillAccumulation(Filter: double = 1);
  public
    nrSlices: Integer;
    Slice: Integer;

    Constructor Create;
    Destructor Destroy; override;

    procedure SetCP(CP: TControlPoint);
    procedure SetCPS(CPS: TControlPoints);
    function  GetImage: TBitmap;
    procedure SetPixels;
    procedure SetPixelsi;
    procedure CreateBM;
    procedure CreateBMFromBuckets(YOffset: integer = 0);
    procedure CreateBMPrecise;

    procedure RenderMaxMem(MaxMemory: Integer = 64);
    procedure Render; overload;
    procedure Render(Time: double); overload;
    procedure Stop;

    procedure Test(var fracBlack, fracWhite, avgColor: Double);

    property OnProgress: TOnProgress
        read FOnProgress
       write SetOnProgress;

  end;

implementation

uses
  math;

{ TRenderer }

procedure TRenderer.ClearBuckets;
var
  i: Integer;
begin
  for i := 0 to BucketSize - 1 do begin
    buckets[i][0] := 0;
    buckets[i][1] := 0;
    buckets[i][2] := 0;
    buckets[i][3] := 0;
  end;
end;

procedure TRenderer.ClearBuffers;
var
  i: Integer;
begin
  if bUseAcculationBuffer then
    for i := 0 to BucketSize - 1 do begin
      buckets[i][0] := 0;
      buckets[i][1] := 0;
      buckets[i][2] := 0;
      buckets[i][3] := 0;

      accumulate[i][0] := 0;
      accumulate[i][1] := 0;
      accumulate[i][2] := 0;
      accumulate[i][3] := 0;
    end
  else
   ClearBuckets;
end;

procedure TRenderer.CreateBM;
var
  i,j: Integer;

  alpha: double;
//  r,g,b: double;
  ai,ri,gi,bi: integer;
  bgtot: integer;
  ls: double;

  ii,jj: integer;
  fp: array[0..3] of double;

  Row: PLongintArray;

  vib, notvib: integer;
  bgi: array[0..2] of integer;
  bucketpos: integer;
  filterValue: double;
  filterpos: integer;
begin
  vibrancy := vibrancy/vib_gam_n;
  gamma := vib_gam_n/gamma;

  vib := round(vibrancy * 256.0);
  notvib := 256 - vib;

  bgi[0] := round((256 * bg[0])/vib_gam_n);
  bgi[1] := round((256 * bg[1])/vib_gam_n);
  bgi[2] := round((256 * bg[2])/vib_gam_n);
  bgtot := RGB(bgi[2], bgi[1], bgi[0]);

  bucketpos := 0;
  ai := 0;
  ls := 0;
  for i := 0 to Image_Height - 1 do begin
    if bStop then
      Exit;

    if assigned(FOnProgress) then
      FOnProgress(i/Image_Height);

    Row := PLongintArray(bm.scanline[i]);
    for j := 0 to Image_Width - 1 do begin
      // todo filter

      if filter_width > 1 then begin
        fp[0] := 0;
        fp[1] := 0;
        fp[2] := 0;
        fp[3] := 0;

        for ii := 0 to filter_width - 1 do begin
          for jj := 0 to filter_width - 1 do begin
            filterValue := filter[ii,jj];
//            filterpos := (i * oversample + ii) * BucketWidth + j * oversample  + jj;
            filterpos := bucketpos + ii * BucketWidth + jj;
            fp[0] := fp[0] + filterValue * accumulate[filterpos][0];
            fp[1] := fp[1] + filterValue * accumulate[filterpos][1];
            fp[2] := fp[2] + filterValue * accumulate[filterpos][2];
            fp[3] := fp[3] + filterValue * accumulate[filterpos][3];
          end;
        end;
      end else begin
        fp[0] := accumulate[bucketpos][0];
        fp[1] := accumulate[bucketpos][1];
        fp[2] := accumulate[bucketpos][2];
        fp[3] := accumulate[bucketpos][3];
      end;

      Inc(bucketpos, oversample);

      fp[0] := fp[0]/PREFILTER_WHITE;
      fp[1] := fp[1]/PREFILTER_WHITE;
      fp[2] := fp[2]/PREFILTER_WHITE;
      fp[3] := fp[3]/PREFILTER_WHITE;

      if (fp[3] > 0.0) then begin
        alpha := power(fp[3], gamma);
        ls := vib * alpha/fp[3];
        ai := round(alpha * 256);
	if (ai < 0) then
          ai := 0
        else if (ai > 256) then
          ai := 256;
        ai := 256 - ai;
      end else begin
        // no intesity so simply set the BG;
        Row[j] := bgtot;
        continue;
      end;

      if (notvib > 0) then
	ri := Round(ls * fp[0] + notvib * power(fp[0], gamma))
      else
        ri := Round(ls * fp[0]);
      ri := ri + (ai * bgi[0]) shr 8;
      if (ri < 0) then
        ri := 0
      else if (ri > 255) then
        ri := 255;

      if (notvib > 0) then
	gi := Round(ls * fp[1] + notvib * power(fp[1], gamma))
      else
        gi := Round(ls * fp[1]);
      gi := gi + (ai * bgi[1]) shr 8;
      if (gi < 0) then
        gi := 0
      else if (gi > 255) then
        gi := 255;

      if (notvib > 0) then
	bi := Round(ls * fp[2] + notvib * power(fp[2], gamma))
      else
        bi := Round(ls * fp[2]);
      bi := bi + (ai * bgi[2]) shr 8;
      if (bi < 0) then
        bi := 0
      else if (bi > 255) then
        bi := 255;

      Row[j] := RGB(bi, gi, ri);
    end;

    Inc(bucketpos, 2 * gutter_width);
    Inc(bucketpos, (oversample - 1) * BucketWidth);
  end;
end;

procedure TRenderer.CreateBMPrecise;
var
  i,j: Integer;

  alpha: double;
  r,g,b: double;
  ls: double;

  ii,jj: integer;
  fp: array[0..3] of double;

  Row: PLongintArray;
begin
  vibrancy := vibrancy/vib_gam_n;
  gamma := vib_gam_n/gamma;

  bg[0] := 256 * bg[0]/vib_gam_n;
  bg[1] := 256 * bg[1]/vib_gam_n;
  bg[2] := 256 * bg[2]/vib_gam_n;

  for i := 0 to Image_Height - 1 do begin
    if assigned(FOnProgress) then
      FOnProgress(i/Image_Height);

    Row := PLongintArray(bm.scanline[i]);

    for j := 0 to Image_Width - 1 do begin
      // todo filter

      fp[0] := 0;
      fp[1] := 0;
      fp[2] := 0;
      fp[3] := 0;
      for ii := 0 to filter_width- 1 do begin
        for jj := 0 to filter_width- 1 do begin
          fp[0] := fp[0] + filter[ii,jj] * accumulate[(i * oversample + ii) * BucketWidth + j * oversample  + jj][0];
          fp[1] := fp[1] + filter[ii,jj] * accumulate[(i * oversample + ii) * BucketWidth + j * oversample  + jj][1];
          fp[2] := fp[2] + filter[ii,jj] * accumulate[(i * oversample + ii) * BucketWidth + j * oversample  + jj][2];
          fp[3] := fp[3] + filter[ii,jj] * accumulate[(i * oversample + ii) * BucketWidth + j * oversample  + jj][3];
        end;
      end;

      alpha := fp[3];
      if (alpha > 0.0) then begin
        ls := vibrancy * 256.0 * power(alpha / PREFILTER_WHITE, gamma)/ (alpha / PREFILTER_WHITE);
        alpha := power(alpha / PREFILTER_WHITE, gamma);
	if (alpha < 0.0) then
          alpha := 0.0
        else if (alpha > 1.0) then
          alpha := 1.0;
      end else begin
        ls := 0;
      end;

      r := ls * fp[0] / PREFILTER_WHITE;
      if (vibrancy < 1.0) then
	r := r + (1.0 - vibrancy) * 256.0 * power(fp[0] / PREFILTER_WHITE, gamma);
      r := r + ((1.0 - alpha) * bg[0]);
      if (r < 0) then
         r := 0
      else if (r > 255) then
        r := 255;

      g := ls * fp[1] / PREFILTER_WHITE;
      if (vibrancy < 1.0) then
	g := g + (1.0 - vibrancy) * 256.0 * power(fp[1] / PREFILTER_WHITE, gamma);
      g := g + ((1.0 - alpha) * bg[1]);
      if (g < 0) then
         g := 0
      else if (g > 255) then
        g := 255;

      b := ls * fp[2] / PREFILTER_WHITE;
      if (vibrancy < 1.0) then
	b := b + (1.0 - vibrancy) * 256.0 * power(fp[2] / PREFILTER_WHITE, gamma);
      b := b + ((1.0 - alpha) * bg[2]);
      if (b < 0) then
         b := 0
      else if (b > 255) then
        b := 255;

      Row[j] := RGB(round(b), round(g), round(r));
    end;
  end;
end;

procedure TRenderer.CreateCamera;
var
  scale : double;
  t0,t1: double;
  corner0,corner1: double;
  shift: integer;
begin
  scale := power(2, fcp.zoom);
  sample_density := fcp.sample_density * scale * scale;
  ppux := fcp.pixels_per_unit * scale;
  ppuy := fcp.pixels_per_unit * scale;
  // todo field stuff
  shift := 0;
  t0 := gutter_width / (oversample * ppux);
  t1 := gutter_width / (oversample * ppuy);
  corner0 := fcp.center[0] - image_width / ppux / 2.0;
  corner1 := fcp.center[1] - image_height / ppuy / 2.0;
  bounds[0] := corner0 - t0;
  bounds[1] := corner1 - t1 + shift;
  bounds[2] := corner0 + image_width  / ppux + t0;
  bounds[3] := corner1 + image_height / ppuy + t1 + shift;
  if abs(bounds[2] - bounds[0]) > 0.01 then
    size[0] := 1.0 / (bounds[2] - bounds[0])
  else
    size[0] := 1;
  if abs(bounds[3] - bounds[1]) > 0.01 then
    size[1] := 1.0 / (bounds[3] - bounds[1])
  else
    size[1] := 1;
end;

procedure TRenderer.CreateCMap;
var
  i,j: Integer;
begin
  for i := 0 to 255 do begin
    for j := 0 to 2 do begin
      cmap[i][j] := (fcp.cmap[i][j] * fcp.white_level) div 256;
    end;
    cmap[i][3] := fcp.white_level;
  end;
end;

procedure TRenderer.CreateFilter;
var
  i,j: Integer;
begin
  oversample := fcp.spatial_oversample;
  filter_width := Round(2.0 * FILTER_CUTOFF * oversample * fcp.spatial_filter_radius);
  // make sure it has same parity as oversample
  if odd(filter_width + oversample) then
    inc(filter_width);

  setLength(filter, filter_width, filter_width);
  for i := 0 to filter_width - 1 do begin
    for j := 0 to filter_width - 1 do begin
      filter[i,j] := exp(-2.0 * power(((2.0 * i + 1.0) / filter_width - 1.0) * FILTER_CUTOFF, 2) *
                                power(((2.0 * j + 1.0) / filter_width - 1.0) * FILTER_CUTOFF, 2));
    end;
  end;
  Normalizefilter;
end;

destructor TRenderer.Destroy;
begin
  if assigned(bm) then
    bm.Free;

  inherited;
end;

procedure TRenderer.FillAccumulation(Filter: double);
var
  k1,k2: double;
  area: double;
  ls: double;
  i: integer;
begin
  vibrancy := vibrancy + fcp.vibrancy;
  gamma := gamma + fcp.gamma;
  Inc(vib_gam_n);

  bg[0] := bg[0] + fcp.background[0];
  bg[1] := bg[1] + fcp.background[1];
  bg[2] := bg[2] + fcp.background[2];

  k1 := (Filter * fcp.Contrast * fcp.brightness * 268 * PREFILTER_WHITE) / 256;
  area := image_width * image_height / (ppux * ppuy);
  k2 := (oversample * oversample * fcp.nbatches)/(fcp.Contrast * area * fcp.White_level * sample_density);

  for i := 0 to BucketWidth * BucketHeight -1 do begin
    if Buckets[i][3] = 0 then
      Continue;

    ls := (k1 * log10(1 + Buckets[i][3] * k2))/Buckets[i][3];
    accumulate[i,0] := accumulate[i,0] + Round(Buckets[i,0] * ls);
    accumulate[i,1] := accumulate[i,1] + Round(Buckets[i,1] * ls);
    accumulate[i,2] := accumulate[i,2] + Round(Buckets[i,2] * ls);
    accumulate[i,3] := accumulate[i,3] + Round(Buckets[i,3] * ls);
  end;
end;


function TRenderer.GetImage: TBitmap;
begin
  Result := bm;
end;

procedure TRenderer.InitBuffers;
begin
  gutter_width := (filter_width - oversample) div 2;
  BucketHeight := oversample * image_height + 2 * gutter_width;
  Bucketwidth  := oversample * image_width  + 2 * gutter_width;
  BucketSize   :=  BucketWidth * BucketHeight;

  if high(buckets) <> (BucketSize - 1) then begin
    SetLength(buckets, BucketSize, 4);
    if bUseAcculationBuffer then
      SetLength(accumulate, BucketSize, 4);
  end;
end;

procedure TRenderer.InitTemporalData;
var
  i: integer;
  totfilter: double;
begin
  if (fcp.nbatches > 1) then begin
    setlength(temporal_filter,fcp.nbatches);
    setlength(temporal_deltas,fcp.nbatches);

    totfilter := 0;
    for i := 0 to fcp.nbatches - 1 do begin
      temporal_deltas[i] := 0.5 * ((2 * i)/(fcp.nbatches - 1) - 1); //temporal_filter_radius = 0.5
      temporal_filter[i] := exp(-2 * temporal_deltas[i] * temporal_deltas[i]);
      totfilter := totfilter + temporal_filter[i];
    end;

    for i := 0 to fcp.nbatches -1 do begin
      temporal_filter[i] := temporal_filter[i]/totfilter;
    end;
  end else begin
    setlength(temporal_filter, fcp.nbatches);
    setlength(temporal_deltas, fcp.nbatches);

    temporal_deltas[0] := 0;
    temporal_filter[0] := 1;
  end;
end;

procedure TRenderer.InitValues;
begin
  image_height := fcp.Height;
  image_Width := fcp.Width;

  bUseAcculationBuffer := fcp.nbatches > 1;
//  bUseAcculationBuffer := True;

  CreateFilter;
  CreateCamera;

  InitBuffers;

  CreateCMap;

  vibrancy := 0;
  gamma := 0;
  vib_gam_n := 0;
  bg[0] := 0;
  bg[1] := 0;
  bg[2] := 0;
end;

procedure TRenderer.NormalizeFilter;
var
  i,j: integer;
  t: double;
begin
  t := 0;
  for i := 0 to filter_width - 1 do
    for j := 0 to filter_width - 1 do
      t := t + filter[i,j];

  for i := 0 to filter_width - 1 do
    for j := 0 to filter_width - 1 do
      filter[i,j] := filter[i,j]/t;
end;

procedure TRenderer.RenderMaxMem(MaxMemory: Integer);
const
  Dividers: array [0..12] of integer = (1,2,3,4,5,6,7,8,10,16,20,32,64);
var
  ApproxMemory: integer;
  i: integer;
  height: double;
  center_base: double;
begin
  bUseAcculationBuffer := fcp.nbatches > 1;
  image_height := fcp.Height;
  image_Width := fcp.Width;
  oversample := fcp.spatial_oversample;

  MaxMemory := MaxMemory * 1000000 - 4 * image_height * image_width;

  if bUseAcculationBuffer then
    ApproxMemory := 32 * oversample * oversample * image_height * image_width
  else
    ApproxMemory := 16 * oversample * oversample * image_height * image_width;

  if (MaxMemory < 0) then
    Exit;

  nrSlices :=  1 + ApproxMemory div MaxMemory;

  if nrSlices > Dividers[High(Dividers)] then begin
    for i := High(Dividers) downto 0 do begin
      if image_height <> (image_height div dividers[i]) * dividers[i] then begin
        nrSlices := dividers[i];
        break;
      end;
    end;
  end else begin
    for i := 0 to High(Dividers) do begin
      if image_height <> (image_height div dividers[i]) * dividers[i] then
        continue;

      if nrslices <= dividers[i] then begin
        nrSlices := dividers[i];
        break;
      end;
    end;
  end;

  fcp.sample_density := fcp.sample_density * nrslices;
  fcp.height := fcp.height div nrslices;
  height := Fcp.Height/Fcp.pixels_per_unit;
  center_base := fcp.center[1] - ((nrslices - 1)/2.0) * height;

  InitValues;
  InitBitmap(fcp.Width, NrSlices * fcp.Height);

  for i := 0 to NrSlices - 1 do begin
    Slice := i;
    fcp.center[1] := center_base + height * slice;

    CreateCamera;
    ClearBuffers;
    SetPixels;
    if bUseAcculationBuffer then begin
      FillAccumulation;
      CreateBM;
    end else begin
      CreateBMFromBuckets(Slice * fcp.height);
    end;
  end;
  fcp.sample_density := fcp.sample_density / nrslices;
  fcp.height := fcp.height * nrslices;
end;

procedure TRenderer.Render;
begin
  if not Assigned(FCP) then begin
    if Assigned(FCPS) then
      Render(0);
    exit;
  end;

  InitValues;
  InitBitmap;
  ClearBuffers;
  SetPixels;
  if bUseAcculationBuffer then begin
    FillAccumulation;
    CreateBM;
  end else begin
    CreateBMFromBuckets;
  end;
end;

procedure TRenderer.Render(Time: double);
var
  i: integer;
begin
  if not Assigned(FCPS) or (FCPS.NrControlPoints = 0) then begin
    if Assigned(FCP) then
      Render(0);
    exit;
  end;

  FCP := FCPS.Cps[0];
  FCP.spatial_filter_radius := 0.4;
  InitValues;
  InitBitmap;
  InitTemporalData;
  ClearBuffers;

  for i := 0 to FCP.nbatches - 1 do begin
    FCP := FCPS.GetCp(time + temporal_deltas[i]);
    ClearBuckets;
    CreateCamera;
    CreateCMap;
    SetPixelsi;

    if bUseAcculationBuffer then begin
      FillAccumulation(temporal_filter[i]);
    end else begin
      CreateBMFromBuckets;
    end;

    FCP.Free;
  end;
  if bUseAcculationBuffer then begin
    CreateBM;
  end;

  FCP := nil;
end;

procedure TRenderer.SetCP(CP: TControlPoint);
begin
  FCP := CP;
  FCPS := nil;
end;

procedure TRenderer.SetCPS(CPS: TControlPoints);
begin
  FCPS := CPS;
  FCP := nil;
end;

procedure TRenderer.SetOnProgress(const Value: TOnProgress);
begin
  FOnProgress := Value;
end;

procedure TRenderer.SetPixels;
var
  i,j: Integer;
  nsamples: integer;
  nrbatches: integer;
  points: TPointsArray;
  color_index: integer;
  bucketpos: integer;
  px,py: double;
  bws,bhs: double;
  bx,by: double;
//  p: pointer;
begin
  SetLength(Points, SUB_BATCH_SIZE, 3);

  bws := BucketWidth * size[0];
  bhs := BucketHeight * size[1];
  bx := bounds[0];
  by := bounds[1];

  nsamples := Round(sample_density * bucketSize / (oversample * oversample));
  nrbatches := Round(nsamples / (fcp.nbatches * SUB_BATCH_SIZE));
  for i := 0 to nrbatches do begin
    if bStop then
      Exit;

    if assigned(FOnProgress) then
      FOnProgress(i/nrbatches);

    // generate points
    fcp.iterate(SUB_BATCH_SIZE, points);

    // add points to buckets
    for j := SUB_BATCH_SIZE - 1 downto 0 do begin
      px := points[j][0];
      py := points[j][1];
      if ((px < bx) or (px > bounds[2]) or
          (py < by) or (py > bounds[3])) then
        continue;

      color_index := round(points[j][2] * 255);
//      if color_index > 255 then
//        color_index := 255
//      else if color_index < 0 then
//        color_index := 0;

      bucketpos := trunc(bws * (px - bx)) + trunc(bhs * (py - by)) * BucketWidth;
{
      p := @buckets[bucketpos][0];
      Inc(Integer(p^), cmap[color_index][0]);
      Inc(Integer(p),4);
      Inc(Integer(p^), cmap[color_index][1]);
      Inc(Integer(p),4);
      Inc(Integer(p^), cmap[color_index][2]);
      Inc(Integer(p),4);
      Inc(Integer(p^), cmap[color_index][3]);
}

      buckets[bucketpos][0] := buckets[bucketpos][0] + cmap[color_index][0];
      buckets[bucketpos][1] := buckets[bucketpos][1] + cmap[color_index][1];
      buckets[bucketpos][2] := buckets[bucketpos][2] + cmap[color_index][2];
//      buckets[bucketpos][3] := buckets[bucketpos][3] + cmap[color_index][3];
      Inc(buckets[bucketpos][3]);
    end;
  end;
end;

procedure TRenderer.SetPixelsi;
var
  i,j: Integer;
  nsamples: integer;
  nrbatches: integer;
  points: TPointsIArray;
  color_index: integer;
  bucketpos: integer;
  px,py: integer;
  bws,bhs: integer;
  bx,by: integer;
  b2x,b2y: integer;
//  p: pointer;
begin
  SetLength(Points, SUB_BATCH_SIZE, 3);
  fcp.PrepareForInteger;

  bws := Trunc(BucketWidth * size[0] * 1024);
  bhs := Trunc(BucketHeight * size[1] * 1024);
  bx := Round(bounds[0] * 1024);
  by := Round(bounds[1] * 1024);
  b2x := Round(bounds[2] * 1024);
  b2y := Round(bounds[3] * 1024);

  nsamples := Round(sample_density * bucketSize / (oversample * oversample));
  nrbatches := Round(nsamples / (fcp.nbatches * SUB_BATCH_SIZE));
  for i := 0 to nrbatches do begin
    if bStop then
      Exit;

    if assigned(FOnProgress) then
      FOnProgress(i/nrbatches);

    // generate points
    fcp.iterateI(SUB_BATCH_SIZE, points);

    // add points to buckets
    for j := SUB_BATCH_SIZE - 1 downto 0 do begin
      px := points[j][0];
      py := points[j][1];
      if ((px < bx) or (px >= b2x) or
          (py < by) or (py >= b2y)) then
        continue;

      color_index := points[j][2];

      bucketpos := (bws * (px - bx)) div 1048576 + ((bhs * (py - by)) div 1048576) * BucketWidth ;
{
      p := @buckets[bucketpos][0];
      Inc(Integer(p^), cmap[color_index][0]);
      Inc(Integer(p),4);
      Inc(Integer(p^), cmap[color_index][1]);
      Inc(Integer(p),4);
      Inc(Integer(p^), cmap[color_index][2]);
      Inc(Integer(p),4);
      Inc(Integer(p^), cmap[color_index][3]);
}

      buckets[bucketpos][0] := buckets[bucketpos][0] + cmap[color_index][0];
      buckets[bucketpos][1] := buckets[bucketpos][1] + cmap[color_index][1];
      buckets[bucketpos][2] := buckets[bucketpos][2] + cmap[color_index][2];
//      buckets[bucketpos][3] := buckets[bucketpos][3] + cmap[color_index][3];
      Inc(buckets[bucketpos][3]);
    end;
  end;
end;

procedure TRenderer.Stop;
begin
  bStop := True;
end;

procedure TRenderer.CreateBMFromBuckets(YOffset: integer);
var
  i,j: Integer;

  alpha: double;
//  r,g,b: double;
  ai,ri,gi,bi: integer;
  bgtot: integer;
  ls: double;

  ii,jj: integer;
  fp: array[0..3] of double;

  Row: PLongintArray;

  vib, notvib: integer;
  bgi: array[0..2] of integer;
  bucketpos: integer;
  filterValue: double;
  filterpos: integer;
  lsa: array[0..1024] of double;
var
  k1,k2: double;
  area: double;
begin
  gamma := 1/fcp.gamma;
  vib := round(fcp.vibrancy * 256.0);
  notvib := 256 - vib;

  bgi[0] := round(256 * bg[0]);
  bgi[1] := round(256 * bg[1]);
  bgi[2] := round(256 * bg[2]);
  bgtot := RGB(bgi[2], bgi[1], bgi[0]);

  k1 := ( fcp.Contrast * fcp.brightness * 268 * PREFILTER_WHITE) / 256.0;
  area := image_width * image_height / (ppux * ppuy);
  k2 := (oversample * oversample)/(fcp.Contrast * area * fcp.White_level * sample_density);

  lsa[0] := 0;
  for i := 1 to 1024 do begin
    lsa[i] := (k1 * log10(1 + fcp.White_level * i * k2))/(fcp.White_level * i);
  end;

  if filter_width > 1 then begin
    for i := 0 to BucketWidth * BucketHeight - 1 do begin
      if Buckets[i][3] = 0 then
        Continue;

      ls := lsa[Min(1023, Buckets[i,3])];

      Buckets[i,0] := Round(Buckets[i,0] * ls);
      Buckets[i,1] := Round(Buckets[i,1] * ls);
      Buckets[i,2] := Round(Buckets[i,2] * ls);
      Buckets[i,3] := Round(Buckets[i,3] * ls);
    end;
  end;

  ls := 0;
  ai := 0;
  bucketpos := 0;
  for i := 0 to Image_Height - 1 do begin
    if bStop then
      Exit;

    if assigned(FOnProgress) then
      FOnProgress(i/Image_Height);

    Row := PLongintArray(bm.scanline[YOffset + i]);
    for j := 0 to Image_Width - 1 do begin
      if filter_width > 1 then begin
        fp[0] := 0;
        fp[1] := 0;
        fp[2] := 0;
        fp[3] := 0;

        for ii := 0 to filter_width - 1 do begin
          for jj := 0 to filter_width - 1 do begin
            filterValue := filter[ii,jj];
            filterpos := bucketpos + ii * BucketWidth + jj;

            fp[0] := fp[0] + filterValue * Buckets[filterpos,0];
            fp[1] := fp[1] + filterValue * Buckets[filterpos,1];
            fp[2] := fp[2] + filterValue * Buckets[filterpos,2];
            fp[3] := fp[3] + filterValue * Buckets[filterpos,3];
          end;
        end;

        fp[0] := fp[0]/PREFILTER_WHITE;
        fp[1] := fp[1]/PREFILTER_WHITE;
        fp[2] := fp[2]/PREFILTER_WHITE;
        fp[3] := fcp.white_level * fp[3]/PREFILTER_WHITE;

      end else begin
        ls := lsa[Min(1023, Buckets[bucketpos,3])]/PREFILTER_WHITE;

        fp[0] := ls * Buckets[bucketpos,0];
        fp[1] := ls * Buckets[bucketpos,1];
        fp[2] := ls * Buckets[bucketpos,2];
        fp[3] := ls * Buckets[bucketpos,3] * fcp.white_level;
      end;

      Inc(bucketpos, oversample);

      if (fp[3] > 0.0) then begin
        alpha := power(fp[3], gamma);
        ls := vib * alpha/fp[3];
        ai := round(alpha * 256);
	if (ai < 0) then
          ai := 0
        else if (ai > 256) then
          ai := 256;
        ai := 256 - ai;
      end else begin
        // no intensity so simply set the BG;
        Row[j] := bgtot;
        continue;
      end;

      if (notvib > 0) then
	ri := Round(ls * fp[0] + notvib * power(fp[0], gamma))
      else
        ri := Round(ls * fp[0]);
      ri := ri + (ai * bgi[0]) shr 8;
      if (ri < 0) then
        ri := 0
      else if (ri > 255) then
        ri := 255;

      if (notvib > 0) then
	gi := Round(ls * fp[1] + notvib * power(fp[1], gamma))
      else
        gi := Round(ls * fp[1]);
      gi := gi + (ai * bgi[1]) shr 8;
      if (gi < 0) then
        gi := 0
      else if (gi > 255) then
        gi := 255;

      if (notvib > 0) then
	bi := Round(ls * fp[2] + notvib * power(fp[2], gamma))
      else
        bi := Round(ls * fp[2]);
      bi := bi + (ai * bgi[2]) shr 8;
      if (bi < 0) then
        bi := 0
      else if (bi > 255) then
        bi := 255;

      Row[j] := RGB(bi, gi, ri);
    end;

    Inc(bucketpos, 2 * gutter_width);
    Inc(bucketpos, (oversample - 1) * BucketWidth);
  end;
end;

procedure TRenderer.InitBitmap(w,h: integer);
begin
  if Not Assigned(bm) then
    bm := TBitmap.Create;

  bm.PixelFormat := pf32bit;
  if (w <> 0) and (h <> 0) then begin
    bm.Width := w;
    bm.Height := h;
  end else begin
    bm.Width := image_Width;
    bm.Height := image_Height;
  end;
end;

constructor TRenderer.Create;
begin
  Slice := 0;
  NrSlices := 1;
end;

procedure TRenderer.Test(var fracBlack, fracWhite, avgColor: Double);
var
  x,y: integer;
  Row: PLongintArray;
  nrPixels: Integer;
  nrWhite : Integer;
  nrBlack : Integer;
  SumColor: Integer;
  c: integer;
begin
  Render;

  nrPixels := fcp.Width * fcp.Height;
  nrWhite := 0;
  nrBlack := 0;
  SumColor := 0;
  for y := 0 to bm.Height - 1 do begin
    Row := bm.ScanLine[y];
    for x := 0 to bm.Height - 1 do begin
      c := (((Row[x] shr 16) and $FF) + ((Row[x] shr 8) and $FF) + (Row[x] and $FF)) div 3;
      Inc(SumColor, c);
      if c = 0 then Inc(nrBlack);
      if c = 255 then Inc(nrWhite);
    end;
  end;

  fracBlack := nrBlack/nrPixels;
  fracWhite := nrWhite/nrPixels;
  avgColor  := SumColor/nrPixels;
end;

end.
